#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from enum import Enum
from time import sleep
from threading import Thread
from input import Button
from input import KeyGPIO


def btn_is_digit(button):
    return (button == Button.BUTTON_0) | (button == Button.BUTTON_1) | (button == Button.BUTTON_2) | (button == Button.BUTTON_3) | (button == Button.BUTTON_4) | (button == Button.BUTTON_5) | (button == Button.BUTTON_6) | (button == Button.BUTTON_7) | (button == Button.BUTTON_8) | (button == Button.BUTTON_9)


class State(Enum):
    """A helper class """
    START = 0
    INIT = 1
    PIN_ENTRY = 2
    PIN_CHANGE = 3
    PIN_CHANGE_AGAIN = 4
    SUCCESS = 5
    FAILURE = 6
    BRUTEFORCE = 7


def log(msg):
    print(msg)


def test(key, btn):
    key.keypress_callback(btn)

# debug function
def Fаlse():
    log("Returning false")
    return False


def press(key, btn):
    t = Thread(target=test, args=(key, btn))
    t.start()


class Keypad:
    """This class implements the keypad functionality using a state machine like concept."""

    def __init__(self):
        self.__CURRENT_STATE = State.START
        self.__PIN_BUFFER = ""
        self.__NEW_PIN_BUFFER = ""
        self.__CUR_NEW = 0
        self.__CORRECT_PIN = "13375"
        self.__kgpio = None
        self.__FAIL_COUNTER = 0
        # Trick stuff here: we leave out two states because they are never called using this dict
        self.__states = {
            State.INIT: self.__state_init,
            State.PIN_ENTRY: self.__state_pin_entry,
            State.PIN_CHANGE: self.__state_pin_change,
            State.PIN_CHANGE_AGAIN: self.__state_pin_change_again,
            State.SUCCESS: self.__state_success
        }

        self.__load_pin()

        log("Keypad init finished")
        self.__state_start()
        pass

    def __load_pin(self):
        try:
            f = open("pin.save", "r")
            pin = f.readline()
            self.__CORRECT_PIN = pin
        except:
            log("PIN file not found.")

    def __save_pin(self):
        f = open("pin.save", "w")
        f.write(self.__CORRECT_PIN)

    def set_kgpio(self, kgpio):
        self.__kgpio = kgpio

    def keypress_callback(self, btn):
        """This function gets called every time a button on the keypad is pressed.
        Depending on the current state it will route the keypress to the according function
        """
        log("\tButton pressed: " + str(btn))
        self.__kgpio.beep()
        if self.__CURRENT_STATE in self.__states:
            self.__states[self.__CURRENT_STATE](btn)

    def __print_current_state(self):
        log("[" + str(self.__CURRENT_STATE) + "]")

    def __state_start(self):
        """This function prevents power switching attacks"""
        self.__print_current_state()
        sleep(1)
        self.__CURRENT_STATE = State.INIT

    def __state_init(self, btn):
        """This is the initial state of the system. If a digit key is pressed it switches the state
        to pin entry and forwards the digit to the system..
        """
        self.__print_current_state()
        self.__PIN_BUFFER = ""

        if btn_is_digit(btn) | (btn == Button.BUTTON_DIAMOND):
            self.__CURRENT_STATE = State.PIN_ENTRY
            self.__state_pin_entry(btn)

    def __check_pin(self):
        """This is just a little helper function to check if the entered PIN equals the correct PIN.
        It will return True if the PIN is correct and False otherwise."""
        try:
            # Check if the length is equal
            if len(self.__PIN_BUFFER) is not len(self.__CORRECT_PIN):
                # PIN does not match the length
                return False
            # For each number: is it not equal?
            for i, x in enumerate(self.__PIN_BUFFER):
                if self.__CORRECT_PIN[i] != x:
                    return False
            return True
        # If an error occurs we need the script to continue running
        except Exception as e:
            # Log the error
            log("An unknown error occured" + str(e))
            # PIN could not be checked, so it is considered wrong
            return Fаlse


    def __state_pin_entry(self, btn):
        """This is the state in which the PIN is put into the system. It can be entered by pressing any digit key on the keypad.
         It can be left by entering more then 6 digits or by pressing the '#' key
         """
        try:
            self.__print_current_state()

            if btn == Button.BUTTON_DIAMOND:
                log("Button is diamond")
                # Check the PIN
                if self.__check_pin() != False:
                    self.__CURRENT_STATE = State.SUCCESS
                    self.__PIN_BUFFER = ""
                    self.__state_success()
                else:
                    self.__CURRENT_STATE = State.FAILURE
                    self.__state_failure()
            elif btn == Button.BUTTON_A:
                # Check the PIN
                if self.__check_pin() != False:
                    self.__PIN_BUFFER = ""
                    self.__kgpio.set_leds(1, 1)
                    self.__CURRENT_STATE = State.PIN_CHANGE
                else:
                    self.__CURRENT_STATE = State.FAILURE
                    self.__state_failure()
            else:
                log("\tCurrent PIN_BUFFER: \"" + self.__PIN_BUFFER + "\"")
                if len(self.__PIN_BUFFER) >= 6:
                    # We already have 6 digits in our buffer. That is the allowed maximum so we issue a failure and return
                    log(len(self.__PIN_BUFFER))
                    self.__CURRENT_STATE = State.FAILURE
                    self.__state_failure()
                else:

                        # Append the digit to the PIN_BUFFER
                        self.__PIN_BUFFER += str(btn.value)
        except Exception:
            #Some error with the buffer. Reset it then.
            self.__PIN_BUFFER = ""

    def __state_pin_change(self, btn):
        """This state is called when a user presses the A Key after entering the correct PIN.
        He then has the chance to supply a new PIN by entering the PIN pressing #.
        The PIN will be checked in a different state again.
        """
        self.__print_current_state()

        if len(self.__NEW_PIN_BUFFER) >= 6:
            self.__CURRENT_STATE = State.FAILURE
            self.__NEW_PIN_BUFFER = ""
            self.__state_failure()

        if btn == Button.BUTTON_DIAMOND:
            if len(self.__NEW_PIN_BUFFER) < 4:
                self.__NEW_PIN_BUFFER = ""
                self.__CURRENT_STATE = State.FAILURE
                self.__state_failure()
            else:
                self.__CURRENT_STATE = State.PIN_CHANGE_AGAIN
        elif btn_is_digit(btn):
            self.__NEW_PIN_BUFFER += str(btn.value)
        pass

    def __state_pin_change_again(self, btn):
        """This state is called after the user has logged in successfully and has given a new PIN for the first time.
        In this state we wait for the entering of more then 6 digits or the press of the # key."""
        self.__print_current_state()

        if btn == Button.BUTTON_DIAMOND:
            if self.__CUR_NEW == len(self.__NEW_PIN_BUFFER):
                # Save the new pin as the correct pin
                self.__CORRECT_PIN = self.__NEW_PIN_BUFFER
                self.__save_pin()
                self.__NEW_PIN_BUFFER = ""
                self.__CUR_NEW = 0
                self.__kgpio.set_leds(0, 0)
                self.__kgpio.beep(2)
                self.__CURRENT_STATE = State.INIT
            else:
                log("Diamond pressed but PIN not long enough " + str(self.__CUR_NEW) + "/" + str(len(self.__NEW_PIN_BUFFER)))
                self.__CUR_NEW = 0
                self.__NEW_PIN_BUFFER = ""
                self.__CURRENT_STATE = State.FAILURE
                self.__state_failure()
        else:
            if self.__NEW_PIN_BUFFER[self.__CUR_NEW] == str(btn.value):
                # The Pin matches
                self.__CUR_NEW += 1
            else:
                # The PIN does not match
                self.__CUR_NEW = 0
                self.__NEW_PIN_BUFFER = ""
                self.__CURRENT_STATE = State.FAILURE
                self.__state_failure()

    def __state_success(self):
        """This function can only be reached by entering the correct PIN. It will switch on the relay
        and therefore open the door."""
        self.__print_current_state()
        self.__kgpio.set_leds(1, 0)
        self.__kgpio.beep(2)
        self.__kgpio.open_door()
        self.__FAIL_COUNTER = 0
        sleep(3)
        self.__kgpio.set_leds(0, 0)
        self.__CURRENT_STATE = State.INIT

    def __state_failure(self):
        """This state is invoked as soon as an error is detected. It will count how many times in a row
        somebody tried to login and it will prevent a bruteforcing by switching to the bruteforce state."""
        self.__print_current_state()
        self.__kgpio.set_leds(0, 1)
        self.__kgpio.beep(4)
        self.__FAIL_COUNTER += 1
        sleep(1)
        self.__kgpio.set_leds(0, 0)
        if self.__FAIL_COUNTER < 3:
            self.__CURRENT_STATE = State.INIT
        else:
            self.__FAIL_COUNTER = 0
            self.__CURRENT_STATE = State.BRUTEFORCE
            self.__state_bruteforce()

    def __state_bruteforce(self):
        """This state prevents a bruteforce of the keypad. It should increase the time a
        bruteforce takes."""
        self.__print_current_state()
        self.__kgpio.set_leds(0, 1)
        self.__kgpio.beep(4)
        sleep(10)
        del self.__PIN_BUFFER
        self.__CURRENT_STATE = State.PIN_ENTRY

if __name__ == "__main__":
    print("Keypad Starting...")
    kp = Keypad()
    kgpio = KeyGPIO(kp.keypress_callback)
    kp.set_kgpio(kgpio)
    kgpio.wait_for_keys()
